(module cpstrans
   (import js-nodes
	   js-parser
	   js-out
	   js-nodes-extra
	   vardecl
	   config
	   verbose
	   walk)
   (export (cps-transform! tree varl)))

(define string_delimiter "\'")

(define *var-list* '())

(define (symbol-to-string-node symbol-arg)
   (let ((str (string-append string_delimiter (string-append (symbol->string symbol-arg) string_delimiter))))
      (make-String str))) 

(define (empty-env)
   (let ((ht (make-hashtable)))
      (hashtable-put! ht 'breaks '())
      (hashtable-put! ht 'continues '())
      (hashtable-put! ht 'return #f)
      (hashtable-put! ht 'label #f)
      ht))

(define (return-env k)
   (let ((ht (make-hashtable)))
      (hashtable-put! ht 'breaks '())
      (hashtable-put! ht 'continues '())
      (hashtable-put! ht 'return k)
      (hashtable-put! ht 'label #f)
      ht))

(define (set-label env label)
   (if (hashtable? env)
       (hashtable-put! env 'label label)
       (error "set-label"
	      "env is NOT a hashtable!" #f)))

(define (take-label env)
   (if (hashtable? env)
       (let ((label (hashtable-get env 'label)))
	  (hashtable-put! env 'label #f)
	  label)
       (error "take-label"
	      "env is NOT a hashtable!" #f)))

(define (push-stack-break env k)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (set! breaks (cons k breaks))
	  (hashtable-put! env 'breaks breaks))))


(define (push-stack-labeled-break env label k)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (set! breaks (cons (cons label k) breaks))
	  (hashtable-put! env 'breaks breaks))))


(define (pop-stack-all-break env)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (set! breaks (cdr breaks))
	      (error "pop-stack-all-break"
		     "No more breaks to pop"
		     #f))
	  (hashtable-put! env 'breaks breaks))))


(define (push-stack-continue env k )
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (set! continues (cons k continues))
	  (hashtable-put! env 'continues continues))))

(define (push-stack-labeled-continue env label k)
   (if (hashtable? env);; (not cps-body) (not block-body) (not (null? bootstrapping-vassigs))
       (let ((continues (hashtable-get env 'continues)))
	  (set! continues (cons (cons label k) continues))
	  (hashtable-put! env 'continues continues))))
   

(define (pop-stack-all-continue env)
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (set! continues (cdr continues))
	      (error "pop-stack-all-continue"
		     "No more continues to pop"
		     #f))
	  (hashtable-put! env 'continues continues))))

(define (search-labeled-break env label)
   (if (hashtable? env)
       (let loop ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (let ((top (car breaks)))
		 (if (and (pair? top)
			  (equal? label (car top)))
		     (cdr top)
		     (loop (cdr breaks))))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (peek-break env)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (let ((top (car breaks)))
		 (if (pair? top)
		     (cdr top)
		     top))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (peek-continue env)
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (let ((top (car continues)))
		 (if (pair? top)
		     (cdr top)
		     top))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (search-labeled-continue env label)
   (if (hashtable? env)
       (let loop ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (let ((top (car continues)))
		 (if (and (pair? top)
			  (equal? label (car top)))
		     (cdr top)
		     (loop (cdr continues))))
	      #f))
       (error "search-labelled-continue" "Not a hashtable!" #f)))
    
(define (get-current-return env)
      (if (hashtable? env)
       (hashtable-get env 'return)
       (error "search-labelled-continue" "Not a hashtable!" #f)))

 
;; Auxiliary Functions for constructing cps terms ========================================================================================================

(define gen-kvar  
   (let ((n 0))
      (lambda ()
         (set! n (+ 1 n))
         (make-Ref (string->symbol (format "_k~a" n))))))

(define gen-var  
   (let ((n 0))
      (lambda ()
         (set! n (+ 1 n))
         (make-Ref (string->symbol (format "_v~a" n))))))

(define dummy-var (gen-var))

(define (extend-var var suffix)  
   (with-access::Ref var (id)
      (let* ((idstr (symbol->string id))
             (newid (string->symbol (format (string-append idstr "_~a") suffix))))
         (make-Ref newid))))

(define (lambda-cps vars exprs) 
   (let ((params (if (list? vars) vars (list vars))) 
         (stmts (if (list? exprs) exprs (list exprs))))
      (make-Cps-Code params (make-Block stmts))))

(define (lambda-cont vars exprs)
   (let ((params (if (list? vars) vars (list vars)))
         (stmts (if (list? exprs) exprs (list exprs))))
      (make-Continu params (make-Block stmts))))

(define (lambda-ncont kvar vars exprs) 
   (let ((params (if (list? vars) vars (list vars)))
         (stmts (if (list? exprs) exprs (list exprs))))
      (make-Named-Continu kvar (make-Continu params (make-Block stmts)))))

(define (apply-cps fun args)
   (let ((largs (if (list? args) args (list args))))
      (make-Call-Cps fun largs)))

(define (apply-cont fun args) ; Generate a call to Continu node.
   (let ((largs (if (list? args) args (list args))))
      (make-Call-Cont fun largs)))

;Javascript Library bindings 
(define FunAccess
   (make-Lib-Binding '__ACCESS))

(define FunFuncall
   (make-Lib-Binding '__FUNCALL))

(define FunAccSig
   (make-Lib-Binding '__ACCSIG))

(define FunBinaryOp
   (make-Lib-Binding '__BINOP))

(define FunUnaryOp 
   (make-Lib-Binding '__UNOP))

(define FunMethodcall
   (make-Lib-Binding '__METHODCALL))

(define FunNew
   (make-Lib-Binding '__NEW))

(define FunGetref
   (make-Lib-Binding '__GETREF))

(define FunMakefun
   (make-Lib-Binding '__MAKEFUN))

(define FunGetProps
   (make-Lib-Binding '__GETPROPS))

(define LocalThis
   (make-Lib-Binding '_this))

(define GetRealValue
   (make-Lib-Binding '_GETREALVALUE))

(define FunFlush
   (make-Lib-Binding '_FLUSH))

;;=========== Main Method ================================================================================================================================
;;========================================================================================================================================================
(define (cps-transform! tree varl)
   (set! *var-list* varl)  
   (cps-trans! tree (empty-env)))
;;========================================================================================================================================================
;;========== Node ========================================================================================================================================
(define-nmethod (Node.cps-trans!)   
   (js-out this (current-error-port) :compress? #f)
   (error "cps-trans!" "Forget node type:" (class-name (object-class this))))

;;;;;;;;;;;;;;;;;;;;;;Statements;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;========== Program =====================================================================================================================================
(define-nmethod (Program.cps-trans!)
   (with-access::Program this (body)
      (default-walk! this)
      this))

(define-nmethod (Var-Program.cps-trans!)
   (with-access::Var-Program this (body var-decl)
      (let ((compiled-body-aux (cps-trans! body env))
            (bootstrapping-code (generate-bootstrapping-code)))
         (let ((bootstrapping-vassigs (cdr bootstrapping-code))
               (bootstrapping-vars (car bootstrapping-code)))
            (let ((new-var-decl 
                     (with-access::Var-Decl-List var-decl (vars)
                        (make-Var-Decl-List (append bootstrapping-vars vars))))
                  (compiled-body 
                     (cond ;; cps-body (not (null? bootstrapping-vassigs))
                           ((and (is-a? compiled-body-aux Cps-Code) (not (null? bootstrapping-vassigs)))
                               (update-cps-term-with-non-cps-code compiled-body-aux bootstrapping-vassigs))
                           ;; cps-body  (null? bootstrapping-vassigs)
                           ((and (is-a? compiled-body-aux Cps-Code) (null? bootstrapping-vassigs))
                               compiled-body-aux)
                           ;; (not cps-body) (not block-body) (not (null? bootstrapping-vassigs))
                           ((and (not (is-a? compiled-body-aux Block)) (not (null? bootstrapping-vassigs)))
                               (make-Block (append bootstrapping-vassigs (list compiled-body-aux))))
                           ;; (not cps-body) (not block-body) (null? bootstrapping-vassigs)
                           ((and (not (is-a? compiled-body-aux Block)) (null? bootstrapping-vassigs))
                               compiled-body-aux)
                           ;; block-body (not (null? bootstrapping-vassigs))
                           ((and (is-a? compiled-body-aux Block) (not (null? bootstrapping-vassigs)))
                               (with-access::Block compiled-body-aux (stmts) 
                                  (make-Block (append bootstrapping-vassigs 
                                     (append stmts (list (make-Call FunFlush '())))))))
                           (else (append compiled-body-aux (list (make-Call FunFlush '())))))))
               (make-Var-Program compiled-body new-var-decl))))))

;;========== BootStrapping Code ===========================================================================================================================
(define (generate-bootstrapping-code) 
   (if (null? *var-list*)
       (cons '() '())
       (let loop ((lst *var-list*)
                  (lst-vassigs '())
                  (lst-vars '()))
          (if (not (null? lst))
             (begin 
                (set! lst-vars (cons (make-Ref (caar lst)) lst-vars))
                (set! lst-vassigs
                   (cons  
                      (make-Vassig 
                         (make-Ref (caar lst))
                         (make-Call FunGetref (list (make-Ref (cadar lst)) (make-Ref (caddar lst)))))
                      lst-vassigs))
                (loop (cdr lst) lst-vassigs lst-vars))
             (cons lst-vars lst-vassigs)))))   

;;========== If  ==========================================================================================================================================
(define (if-cps-transform env test then else) 
   (let* ((k (gen-kvar))
          (v (gen-var))
          (compiled-test (cps-trans! test env))
          (compiled-then (cps-trans! then env))
          (compiled-else (cps-trans! else env))
          (if-cont-var (extend-var k "if_body_cont")))
      (let ((if-cont 
               (lambda-cont v 
                   (make-If v 
                      (if (is-a? compiled-then Cps-Code)
                         (apply-cps compiled-then k)
                         (my-make-block 
                            (list 
                               compiled-then
                               (make-Call k (list (make-Undefined 0)))))) 
                      (if (is-a? compiled-else Cps-Code)
                         (apply-cps compiled-else k)
                         (my-make-block 
                            (list 
                               compiled-else
                               (make-Call k (list (make-Undefined 0))))))))))
          (lambda-cps k
             (list 
                (make-Var-Decl-List
                   (list (make-Init if-cont-var if-cont)))
                (if (is-a? compiled-test Cps-Code) 
                   (apply-cps compiled-test 
                      (lambda-cont v
                         (make-Call GetRealValue (list v if-cont-var))))    
                (make-Call GetRealValue (list compiled-test if-cont-var)))))))) 

(define-nmethod (If.cps-trans!) 
   (with-access::If this (test then else)
      (if-cps-transform env test then else)))

(define (my-make-block lst-exprs)
   (let ((ret-block (make-Block (merge-consecutive-blocks lst-exprs)))) 
     ret-block))

;;========== For  =========================================================================================================================================
(define-nmethod (For.cps-trans!)
   (with-access::For this (init test incr body)
      (let ((k (gen-kvar)))
	 (if (not init)
	     (set! init (make-Bool "true")))
	 (if (not test)
	     (set! test (make-Bool "true")))
	 (if (not incr)
	     (set! incr (make-Bool "true")))
	 (lambda-cps k 
            (make-for-cont-aux env init test incr body k)))))

(define (make-for-cont-aux env init test incr body k)
   (let ((compiled-init (cps-trans! init env)))
      (let ((label (take-label env))
            (body-cont-var (extend-var k "for_body_cont"))
            (test-cont-var (extend-var k "for_test_cont"))
            (incr-cont-var (extend-var k "for_incr_cont"))
            (v (gen-var)))
         (if label
            (begin 
               (push-stack-labeled-break env label k)
               (push-stack-labeled-continue env label incr-cont-var))
            (begin
               (push-stack-break env k)
               (push-stack-continue env incr-cont-var)))
         (let ((compiled-test (cps-trans! test env))
               (compiled-body (cps-trans! body env))
               (compiled-incr (cps-trans! incr env)))
            (let ((test-cont 
                     (if (is-a? compiled-test Cps-Code)
                        (lambda-cont dummy-var
                           (apply-cps compiled-test 
                              (lambda-cont v (make-Call GetRealValue (list v body-cont-var)))))
                        (lambda-cont dummy-var
                           (list 
                              (make-Var-Decl-List (list (make-Init v compiled-test)))
                              (make-Call GetRealValue (list v body-cont-var))))))
                  (body-cont
                     (lambda-cont v
                        (make-If v
                           (if (is-a? compiled-body Cps-Code)
                              (apply-cps compiled-body incr-cont-var)
                              (my-make-block 
                                 (list 
                                    compiled-body
                                    (make-Call incr-cont-var (list (make-Undefined 0))))))
                           (apply-cont k (make-Undefined 0)))))
                  (incr-cont 
                     (if (is-a? compiled-incr Cps-Code)
                        (lambda-cont dummy-var
                           (apply-cps compiled-incr test-cont-var))
                        (lambda-cont dummy-var 
                           (list 
                              compiled-incr 
                              (make-Call test-cont-var (list (make-Undefined 0))))))))
               (pop-stack-all-break env)
               (pop-stack-all-continue env)
               (make-for-cont-aux-aux compiled-init test-cont test-cont-var incr-cont incr-cont-var body-cont body-cont-var)))))) 
 
(define (make-for-cont-aux-aux compiled-init test-cont test-cont-var incr-cont incr-cont-var body-cont body-cont-var)
   (let ((main-code 
            (list 
               (make-Var-Decl-List (list (make-Init test-cont-var test-cont)))
               (make-Var-Decl-List (list (make-Init incr-cont-var incr-cont)))
               (make-Var-Decl-List (list (make-Init body-cont-var body-cont)))
               (make-Call test-cont-var (list (make-Undefined 0))))))
      (if (is-a? compiled-init Cps-Code)
         (apply-cps compiled-init 
            (lambda-cont dummy-var 
               main-code))
         (cons compiled-init main-code)))) 

;;========== While  =======================================================================================================================================
(define-nmethod (While.cps-trans!) 
   (with-access::While this (test body)
       (let* ((k (gen-kvar))
              (v (gen-var))
              (body-cont-var (extend-var k "while_body_cont"))
              (test-cont-var (extend-var k "while_test_cont")))
          (let ((label (take-label env))) 
             (if label
                (begin
                   (push-stack-labeled-break env label k)
                   (push-stack-labeled-continue env label test-cont-var))
                (begin
                   (push-stack-break env k)
                   (push-stack-continue env test-cont-var))))
          (let ((compiled-test (cps-trans! test env))
                (compiled-body (cps-trans! body env))) 
             (let ((test-cont 
                      (if (is-a? compiled-test Cps-Code)
                         (lambda-cont dummy-var
                            (apply-cps compiled-test 
                               (lambda-cont v
                                  (make-Call GetRealValue (list v body-cont-var)))))
                         (lambda-cont dummy-var
                            (make-Call GetRealValue (list compiled-test body-cont-var)))))
                   (body-cont 
                      (lambda-cont v
                         (make-If v 
                            (if (is-a? compiled-body Cps-Code)
                               (apply-cps compiled-body test-cont-var)
                               (my-make-block 
                                  (list 
                                     compiled-body
                                     (make-Call test-cont-var (list (make-Undefined 0))))))
                            (apply-cont k (make-Undefined 0))))))
                (let ((cps-code 
                         (lambda-cps k 
                            (my-make-block     
                               (list 
                                  (make-Var-Decl-List 
                                     (list (make-Init test-cont-var test-cont)))
                                  (make-Var-Decl-List 
                                     (list (make-Init body-cont-var body-cont)))
                                  (make-Call test-cont-var (list (make-Undefined 0))))))))
                   (pop-stack-all-break env)
                   (pop-stack-all-continue env)
                   cps-code))))))

;;========== Do ===========================================================================================================================================
(define-nmethod (Do.cps-trans!)
   (with-access::Do this (body test)
      (let* ((k (gen-kvar))
             (v (gen-var))
             (k-while (gen-kvar))
             (body-cont-var (extend-var k "do_body_cont"))
             (test-cont-var (extend-var k "do_test_cont")))
         (let ((label (take-label env))) 
            (if label
               (begin
                  (push-stack-labeled-break env label k)
                  (push-stack-labeled-continue env label test-cont-var))
               (begin
                  (push-stack-break env k)
                  (push-stack-continue env test-cont-var))))
         (let ((compiled-test (cps-trans! test env))
               (compiled-body (cps-trans! body env)))   
            (let ((test-cont 
                     (if (is-a? compiled-test Cps-Code)
                        (lambda-cont dummy-var
                           (apply-cps compiled-test 
                              (lambda-cont v
                                 (make-Call GetRealValue (list v body-cont-var)))))
                        (lambda-cont dummy-var
                           (make-Call GetRealValue (list compiled-test body-cont-var)))))
                  (body-cont 
                     (lambda-cont v
                        (make-If v 
                           (if (is-a? compiled-body Cps-Code)
                              (apply-cps compiled-body test-cont-var)
                              (my-make-block 
                                 (list 
                                    compiled-body
                                    (make-Call test-cont-var (list (make-Undefined 0))))))
                           (apply-cont k (make-Undefined 0))))))
               (let ((cps-code 
                         (lambda-cps k 
                            (list 
                               (make-Var-Decl-List 
                                  (list 
                                     (make-Init test-cont-var test-cont)))
                               (make-Var-Decl-List 
                                  (list 
                                     (make-Init body-cont-var body-cont)))
                               (if (is-a? compiled-body Cps-Code)
                                  (make-Call compiled-body (list test-cont-var))
                                  (my-make-block 
                                     (list 
                                        compiled-body
                                        (make-Call test-cont-var (list (make-Undefined 0))))))))))
                  (pop-stack-all-break env)
	          (pop-stack-all-continue env)
	          cps-code)))))) 

;;========== Continue  ===================================================================================================================================
(define-nmethod (Continue.cps-trans!)
   (with-access::Continue this (id)
      (let ((k (gen-kvar))
            (k-continue 
               (if (not id)
                  (peek-continue env)
                  (search-labeled-continue env id))))
         (if k-continue
            (lambda-cps k 
               (apply-cont k-continue (make-Undefined 0)))
            (error "cps-trans!" "Invalid continue statement!" (if id id #f))))))

;;========== Break  ====================================================================================================================================== 
(define-nmethod (Break.cps-trans!)
   (with-access::Break this (id)
      (let ((k (gen-kvar))
            (k-break 
               (if (not id)
                  (peek-break env)
                  (search-labeled-break env id))))
         (if k-break
            (lambda-cps k 
               (apply-cont k-break (make-Undefined 0)))
            (error "cps-trans!" "Invalid break statement!" (if id id #f))))))
   
;;========== Return  =====================================================================================================================================
(define-nmethod (Return.cps-trans!)
   (with-access::Return this (val)
      (let ((k (gen-kvar))
            (compiled-ret-val (cps-trans! val env))
            (return-cont (get-current-return env)))
         (if return-cont
            (if (is-a? compiled-ret-val Cps-Code)
                (lambda-cps k 
                   (apply-cps compiled-ret-val return-cont))
                (my-make-block
                   (list 
                      (apply-cont return-cont compiled-ret-val)
                      (make-Return (make-Undefined 0)))))
            (error "cps-trans!" "no return point" #f)))))

;;========== With ========================================================================================================================================
(define-nmethod (With.cps-trans!) 
  (error "cps-trans!" "With statement not supported yet!"
     (class-name (object-class this))))

;;========== Labeled =====================================================================================================================================
(define-nmethod (Labeled.cps-trans!)
   (with-access::Labeled this (id body)
      (if (or (For? body) (While? body) (Do? body))
         (begin
            (set-label env id)
            (cps-trans! body env))
         (cps-trans! body env))))

;;========== Vassig  =====================================================================================================================================
(define-nmethod (Vassig.cps-trans!) 
   (with-access::Vassig this (lhs rhs)
      (let ((k (gen-kvar))
            (v-rhs (extend-var (gen-var) "rhs"))
            (compiled-rhs (cps-trans! rhs env)))
        (if (is-a? compiled-rhs Cps-Code)
           (lambda-cps k
              (apply-cps compiled-rhs
                 (lambda-cont v-rhs
                    (list
                       (make-Vassig lhs v-rhs)
                       (apply-cont k v-rhs)))))
           (if (is-a? compiled-rhs Block)
              (block-to-Vassig lhs compiled-rhs)     
              (make-Vassig lhs compiled-rhs))))))

(define (last-element lst)
   (cond 
      ((null? lst) lst)
      ((null? (cdr lst)) (car lst))
      (else (last-element (cdr lst)))))

(define (remove-last-element lst)
   (if (null? lst) 
      lst
      (let loop ((lst-aux lst)
                 (lst-ret (list)))
         (if (null? (cdr lst-aux))
            (reverse lst-ret)
            (loop (cdr lst-aux) (cons (car lst-aux) lst-ret)))))) 

(define (block-to-Vassig lhs compiled-rhs) 
   (with-access::Block compiled-rhs (stmts) 
      (let ((last-expr (last-element stmts))
            (stmts-without-last (remove-last-element stmts)))
         (let ((vassig-expr (make-Vassig lhs last-expr)))
            (make-Block (append stmts-without-last (list vassig-expr)))))))

;;========== Accsig  =====================================================================================================================================
(define (convert-simple-expr-to-cps expr)
  (if (is-a? expr Cps-Code) 
     expr
     (let ((k (gen-kvar)))
        (lambda-cps k 
           (apply-cont k expr)))))

(define-nmethod (Accsig.cps-trans!)
   (with-access::Accsig this (lhs rhs)
      (with-access::Access lhs (obj field)
         (let ((compiled-object (cps-trans! obj env)) 
               (compiled-field (cps-trans! field env)) 
               (compiled-rhs (cps-trans! rhs env)))
            (let ((is-object-cps (is-a? compiled-object Cps-Code))
                  (is-field-cps  (is-a? compiled-field Cps-Code))
                  (is-rhs-cps (is-a? compiled-rhs Cps-Code)))
               (cond 
                  ;; cps-object cps-field cps-rhs
                  ((and is-object-cps is-field-cps is-rhs-cps) (compile-accsig-aux1 compiled-object compiled-field compiled-rhs))
                  ;; cps-object cps-field (not cps-rhs)
                  ((and is-object-cps is-field-cps (not is-rhs-cps)) (compile-accsig-aux2 compiled-object compiled-field compiled-rhs))
                  ;; cps-object (not cps-field) cps-rhs
                  ((and is-object-cps (not is-field-cps) is-rhs-cps) (compile-accsig-aux3 compiled-object compiled-field compiled-rhs))
                  ;; cps-object (not cps-field) (not cps-rhs)
                  ((and is-object-cps (not is-field-cps) (not is-rhs-cps)) (compile-accsig-aux4 compiled-object compiled-field compiled-rhs))
                  ;; (not cps-object) cps-field cps-rhs
                  ((and (not is-object-cps) is-field-cps is-rhs-cps) (compile-accsig-aux5 compiled-object compiled-field compiled-rhs))
                  ;; (not cps-object) cps-field (not cps-rhs)
                  ((and (not is-object-cps) is-field-cps (not is-rhs-cps)) (compile-accsig-aux6 compiled-object compiled-field compiled-rhs))
                  ;; (not cps-object) (not cps-field) cps-rhs
                  ((and (not is-object-cps) (not is-field-cps) is-rhs-cps) (compile-accsig-aux7 compiled-object compiled-field compiled-rhs))
                  ;; (not cps-object) (not cps-field) (not cps-rhs)
                  (else (compile-accsig-aux8 compiled-object compiled-field compiled-rhs))))))))

;; 1 - cps-object cps-field cps-rhs                  
(define (compile-accsig-aux1 cps-compiled-object cps-compiled-field cps-compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-field (extend-var v "field"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-field 
                     (lambda-cont v-field
                        (apply-cps cps-compiled-rhs
                           (lambda-cont v-rhs
                              (apply-cont k (make-Call FunAccSig (list v-object v-field v-rhs)))))))))))))

;; 2 - cps-object cps-field (not cps-rhs)
(define (compile-accsig-aux2 cps-compiled-object cps-compiled-field compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-field 
                     (lambda-cont v-field
                        (apply-cont k (make-Call FunAccSig (list v-object v-field compiled-rhs)))))))))))

;; 3 - cps-object (not cps-field) cps-rhs                
(define (compile-accsig-aux3 cps-compiled-object compiled-field cps-compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-rhs
                     (lambda-cont v-rhs
                        (apply-cont k (make-Call FunAccSig (list v-object compiled-field v-rhs)))))))))))

;; 4 - cps-object (not cps-field) (not cps-rhs)
(define (compile-accsig-aux4 cps-compiled-object compiled-field compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cont k (make-Call FunAccSig (list v-object compiled-field compiled-rhs)))))))))

;; 5 - (not cps-object) cps-field cps-rhs
(define (compile-accsig-aux5 compiled-object cps-compiled-field cps-compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-field (extend-var v "field"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-field 
               (lambda-cont v-field
                  (apply-cps cps-compiled-rhs
                     (lambda-cont v-rhs
                        (apply-cont k (make-Call FunAccSig (list compiled-object v-field v-rhs)))))))))))

;; 6 - (not cps-object) cps-field (not cps-rhs)
(define (compile-accsig-aux6 compiled-object cps-compiled-field compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-field 
               (lambda-cont v-field
                  (apply-cont k (make-Call FunAccSig (list compiled-object v-field compiled-rhs)))))))))

;; 7 - (not cps-object) (not cps-field) cps-rhs
(define (compile-accsig-aux7 compiled-object compiled-field cps-compiled-rhs)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-rhs
               (lambda-cont v-rhs
                  (apply-cont k (make-Call FunAccSig (list compiled-object compiled-field v-rhs)))))))))

;; 8 - (not cps-object) (not cps-field) (not cps-rhs)
(define (compile-accsig-aux8 compiled-object compiled-field compiled-rhs)
   (make-Call FunAccSig (list compiled-object compiled-field compiled-rhs)))   

;;========== Vassig-op  ==================================================================================================================================
;; Nao estou a prever o caso: x+=(4,5,6); (isto e uma estupidez!!!)
(define-nmethod (Vassig-op.cps-trans!) 
   (with-access::Vassig-op this (lhs rhs op)
      (let ((k (gen-kvar))
            (v-rhs (extend-var (gen-var) "rhs"))
            (compiled-rhs (cps-trans! rhs env)))
         (if (is-a? compiled-rhs Cps-Code)
            (lambda-cps k
               (apply-cps compiled-rhs
                  (lambda-cont v-rhs
                     (list
                        (make-Vassig 
                           lhs 
                           (make-Call FunBinaryOp (list (symbol-to-string-node op) lhs v-rhs))) 
                        (apply-cont k lhs)))))
            (make-Vassig 
               lhs 
               (make-Call FunBinaryOp (list (symbol-to-string-node op) lhs compiled-rhs)))))))

;;========== Accsig-op  ==================================================================================================================================
(define-nmethod (Accsig-op.cps-trans!)
   (with-access::Accsig-op this (lhs rhs op)
      (with-access::Access lhs (obj field)
         (let ((compiled-object (cps-trans! obj env)) 
               (compiled-field (cps-trans! field env)) 
               (compiled-rhs (cps-trans! rhs env)))
            (let ((is-object-cps (is-a? compiled-object Cps-Code))
                  (is-field-cps  (is-a? compiled-field Cps-Code))
                  (is-rhs-cps (is-a? compiled-rhs Cps-Code))
                  (op-string (symbol-to-string-node op)))
               (cond 
                  ;; cps-object cps-field cps-rhs
                  ((and is-object-cps is-field-cps is-rhs-cps) (compile-accsigop-aux1 compiled-object compiled-field compiled-rhs op-string))
                  ;; cps-object cps-field (not cps-rhs)
                  ((and is-object-cps is-field-cps (not is-rhs-cps)) (compile-accsigop-aux2 compiled-object compiled-field compiled-rhs op-string))
                  ;; cps-object (not cps-field) cps-rhs
                  ((and is-object-cps (not is-field-cps) is-rhs-cps) (compile-accsigop-aux3 compiled-object compiled-field compiled-rhs op-string))
                  ;; cps-object (not cps-field) (not cps-rhs)
                  ((and is-object-cps (not is-field-cps) (not is-rhs-cps)) (compile-accsigop-aux4 compiled-object compiled-field compiled-rhs op-string))
                  ;; (not cps-object) cps-field cps-rhs
                  ((and (not is-object-cps) is-field-cps is-rhs-cps) (compile-accsigop-aux5 compiled-object compiled-field compiled-rhs op-string))
                  ;; (not cps-object) cps-field (not cps-rhs)
                  ((and (not is-object-cps) is-field-cps (not is-rhs-cps)) (compile-accsigop-aux6 compiled-object compiled-field compiled-rhs op-string))
                  ;; (not cps-object) (not cps-field) cps-rhs
                  ((and (not is-object-cps) (not is-field-cps) is-rhs-cps) (compile-accsigop-aux7 compiled-object compiled-field compiled-rhs op-string))
                  ;; (not cps-object) (not cps-field) (not cps-rhs)
                  (else (compile-accsigop-aux8 compiled-object compiled-field compiled-rhs op-string))))))))

;; 1 - cps-object cps-field cps-rhs                  
(define (compile-accsigop-aux1 cps-compiled-object cps-compiled-field cps-compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-field (extend-var v "field"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-field 
                     (lambda-cont v-field
                        (apply-cps cps-compiled-rhs
                           (lambda-cont v-rhs
                              (apply-cont k 
                                 (make-Call FunAccSig 
                                    (list v-object v-field
                                       (make-Call FunBinaryOp
                                          (list op-string 
                                             (make-Call FunAccess (list v-object v-field))
                                             v-rhs)))))))))))))))

;; 2 - cps-object cps-field (not cps-rhs)
(define (compile-accsigop-aux2 cps-compiled-object cps-compiled-field compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-field 
                     (lambda-cont v-field
                        (apply-cont k 
                           (make-Call FunAccSig 
                              (list v-object v-field
                                 (make-Call FunBinaryOp
                                    (list op-string 
                                       (make-Call FunAccess (list v-object v-field))
                                       compiled-rhs)))))))))))))
;; 3 - cps-object (not cps-field) cps-rhs                
(define (compile-accsigop-aux3 cps-compiled-object compiled-field cps-compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-rhs
                     (lambda-cont v-rhs
                        (apply-cont k 
                           (make-Call FunAccSig 
                              (list v-object compiled-field
                                 (make-Call FunBinaryOp
                                    (list op-string 
                                       (make-Call FunAccess (list v-object compiled-field))
                                       v-rhs)))))))))))))

;; 4 - cps-object (not cps-field) (not cps-rhs)
(define (compile-accsigop-aux4 cps-compiled-object compiled-field compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cont k 
                     (make-Call FunAccSig 
                        (list v-object compiled-field
                           (make-Call FunBinaryOp
                              (list op-string 
                                 (make-Call FunAccess (list v-object compiled-field))
                                 compiled-rhs)))))))))))

;; 5 - (not cps-object) cps-field cps-rhs
(define (compile-accsigop-aux5 compiled-object cps-compiled-field cps-compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-field (extend-var v "field"))
            (v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-field 
               (lambda-cont v-field
                  (apply-cps cps-compiled-rhs
                     (lambda-cont v-rhs
                        (apply-cont k 
                           (make-Call FunAccSig 
                              (list compiled-object v-field
                                 (make-Call FunBinaryOp
                                    (list op-string 
                                       (make-Call FunAccess (list compiled-object v-field))
                                       v-rhs)))))))))))))

;; 6 - (not cps-object) cps-field (not cps-rhs)
(define (compile-accsigop-aux6 compiled-object cps-compiled-field compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-field 
               (lambda-cont v-field
                  (apply-cont k 
                     (make-Call FunAccSig 
                        (list compiled-object v-field
                           (make-Call FunBinaryOp
                              (list op-string 
                                 (make-Call FunAccess (list compiled-object v-field))
                                 compiled-rhs)))))))))))

;; 7 - (not cps-object) (not cps-field) cps-rhs
(define (compile-accsigop-aux7 compiled-object compiled-field cps-compiled-rhs op-string)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-rhs (extend-var v "rhs")))
         (lambda-cps k
            (apply-cps cps-compiled-rhs
               (lambda-cont v-rhs
                  (apply-cont k 
                     (make-Call FunAccSig 
                        (list compiled-object compiled-field
                           (make-Call FunBinaryOp
                              (list op-string 
                                 (make-Call FunAccess (list compiled-object compiled-field))
                                 v-rhs)))))))))))

;; 8 - (not cps-object) (not cps-field) (not cps-rhs)
(define (compile-accsigop-aux8 compiled-object compiled-field compiled-rhs op-string)
   (make-Call FunAccSig 
      (list compiled-object compiled-field
         (make-Call FunBinaryOp
            (list op-string 
               (make-Call FunAccess (list compiled-object compiled-field))
               compiled-rhs)))))

;;========== Binary  =====================================================================================================================================
(define-nmethod (Binary.cps-trans!)
   (with-access::Binary this (lhs op rhs)
      (let ((compiled-lhs (cps-trans! lhs env))
            (compiled-rhs (cps-trans! rhs env)))
         (let ((is-lhs-cps (is-a? compiled-lhs Cps-Code))
               (is-rhs-cps (is-a? compiled-rhs Cps-Code)))
            (cond 
               ;; cps-lhs cps-rhs
               ((and is-lhs-cps is-rhs-cps) (compile-binary-aux1 compiled-lhs compiled-rhs op))
               ;; cps-lhs (not cps-rhs)
               ((and is-lhs-cps (not is-rhs-cps)) (compile-binary-aux2 compiled-lhs compiled-rhs op))
               ;; (not cps-lhs) cps-rhs
               ((and (not is-lhs-cps) is-rhs-cps) (compile-binary-aux3 compiled-lhs compiled-rhs op))
               ;; (not cps-lhs) (not cps-rhs)
               (else (compile-binary-aux4 compiled-lhs compiled-rhs op)))))))

;; cps-lhs cps-rhs
(define (compile-binary-aux1 cps-compiled-lhs cps-compiled-rhs op)
  (let ((k (gen-kvar))
        (v (gen-var)))
     (let ((vl (extend-var v "lhs"))
           (vr (extend-var v "rhs")))
        (lambda-cps k 
           (apply-cps cps-compiled-lhs
              (lambda-cont vl 
                 (apply-cps cps-compiled-rhs
                    (lambda-cont vr
                       (apply-cont k 
                          (make-Call FunBinaryOp 
                             (list (symbol-to-string-node op) vl vr)))))))))))

;; cps-lhs (not cps-rhs)            
(define (compile-binary-aux2 cps-compiled-lhs compiled-rhs op)
  (let ((k (gen-kvar))
        (v (gen-var)))
     (let ((vl (extend-var v "lhs")))
        (lambda-cps k 
           (apply-cps cps-compiled-lhs
              (lambda-cont vl 
                 (apply-cont k 
                    (make-Call FunBinaryOp 
                       (list (symbol-to-string-node op) vl compiled-rhs)))))))))

;; (not cps-lhs) cps-rhs
(define (compile-binary-aux3 compiled-lhs cps-compiled-rhs op)
  (let ((k (gen-kvar))
        (v (gen-var)))
     (let ((vr (extend-var v "rhs")))
        (lambda-cps k 
           (apply-cps cps-compiled-rhs
              (lambda-cont vr
                 (apply-cont k 
                     (make-Call FunBinaryOp 
                         (list (symbol-to-string-node op) compiled-lhs vr)))))))))

;; (not cps-lhs) (not cps-rhs)
(define (compile-binary-aux4 compiled-lhs compiled-rhs op)
   (make-Call FunBinaryOp 
      (list (symbol-to-string-node op) compiled-lhs compiled-rhs)))      
                  
;;========== Unary  ======================================================================================================================================
(define-nmethod (Unary.cps-trans!) 
   (with-access::Unary this (op expr)
      (let ((compiled-expr (cps-trans! expr env)))
         (if (is-a? compiled-expr Cps-Code)
            (let ((k (gen-kvar))
                  (v (gen-var)))
               (lambda-cps k
                  (apply-cps compiled-expr
                     (lambda-cont v
                        (apply-cont k 
                           (make-Call FunUnaryOp (list op v)))))))
             (make-Call FunUnaryOp (list (symbol-to-string-node op) compiled-expr))))))

;;========== Access  =====================================================================================================================================
(define-nmethod (Access.cps-trans!) 
   (with-access::Access this (obj field)
      (let ((compiled-obj (cps-trans! obj env))
            (compiled-field (cps-trans! field env)))
         (let ((is-obj-cps (is-a? compiled-obj Cps-Code))
               (is-field-cps (is-a? compiled-field Cps-Code)))
            (cond 
               ;; cps-ojbect cps-field
               ((and is-obj-cps is-field-cps) (compile-access-aux1 compiled-obj compiled-field))
               ;; cps-object (not cps-field)
               ((and is-obj-cps (not is-field-cps)) (compile-access-aux2 compiled-obj compiled-field))
               ;; (not cps-ojbect) cps-field
               ((and (not is-obj-cps) is-field-cps) (compile-access-aux3 compiled-obj compiled-field))
               ;; (not cps-ojbect) (not cps-field)
               (else (compile-access-aux4 compiled-obj compiled-field)))))))
         
;; 1 - cps-ojbect cps-field
(define (compile-access-aux1 cps-compiled-object cps-compiled-field)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object"))
            (v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cps cps-compiled-field
                     (lambda-cont v-field
                        (apply-cont k 
                           (make-Call FunAccess (list v-object v-field)))))))))))   
            
;; 2 - cps-ojbect (not cps-field)
(define (compile-access-aux2 cps-compiled-object compiled-field)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-object (extend-var v "object")))
         (lambda-cps k
            (apply-cps cps-compiled-object
               (lambda-cont v-object
                  (apply-cont k 
                     (make-Call FunAccess (list v-object compiled-field)))))))))

;; 3 - (not cps-ojbect) cps-field
(define (compile-access-aux3 compiled-object cps-compiled-field)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-field (extend-var v "field")))
         (lambda-cps k
            (apply-cps cps-compiled-field
               (lambda-cont v-field
                  (apply-cont k 
                     (make-Call FunAccess (list compiled-object v-field)))))))))

;; 4 - (not cps-ojbect) (not cps-field)
(define (compile-access-aux4 compiled-object compiled-field)
   (make-Call FunAccess (list compiled-object compiled-field)))

;;========== Postfix =====================================================================================================================================
(define-nmethod (Postfix.cps-trans!)
   (with-access::Postfix this (expr op)
      (cond ((Access? expr)
               (with-access::Access expr (obj field)
                  (compile-postfixed-op-access env obj field op)))
            ((Ref? expr) 
               (compile-postfixed-op-ref env expr op))
            (else
               (error "cps-trans!" "Missing case in Postfix.cps-trans!" #f))))) 

(define (compile-postfixed-op-access env obj field op)
   (let ((compiled-object (cps-trans! obj env))
         (compiled-field (cps-trans! field env)))
      (let ((is-cps-object (is-a? compiled-object Cps-Code))
            (is-cps-field (is-a? compiled-field Cps-Code))) 
         (cond 
            ;; cps-ojbect cps-field
            ((and is-cps-object is-cps-field) (compile-postfixed-op-access-aux1 compiled-object compiled-field op))
            ;; cps-object (not cps-field)
            ((and is-cps-object (not is-cps-field)) (compile-postfixed-op-access-aux2 compiled-object compiled-field op))
            ;; (not cps-ojbect) cps-field
            ((and (not is-cps-object) is-cps-field) (compile-postfixed-op-access-aux3 compiled-object compiled-field op))
            ;; (not cps-ojbect) (not cps-field)
            (else (compile-postfixed-op-access-aux4 compiled-object compiled-field op))))))

;; 1 - cps-ojbect cps-field
(define (compile-postfixed-op-access-aux1 cps-compiled-object cps-compiled-field op)
   (let ((k (gen-kvar))
         (v-obj (gen-var))
         (v-field (gen-var)))
      (lambda-cps k 
         (apply-cps cps-compiled-object 
            (lambda-cont v-obj 
               (apply-cps cps-compiled-field
                  (lambda-cont v-field
                     (apply-cont k 
                        (make-Call FunUnaryOp 
                           (list (symbol-to-string-node op)
                              (make-Call FunAccess (list v-obj v-field))))))))))))

;; 2 - cps-object (not cps-field)
(define (compile-postfixed-op-access-aux2 cps-compiled-object compiled-field op)
   (let ((k (gen-kvar))
         (v-obj (gen-var)))
      (lambda-cps k 
         (apply-cps cps-compiled-object 
            (lambda-cont v-obj 
               (apply-cont k 
                  (make-Call FunUnaryOp 
                     (list (symbol-to-string-node op)
                        (make-Call FunAccess (list v-obj compiled-field))))))))))

;; 3 - (not cps-ojbect) cps-field 
(define (compile-postfixed-op-access-aux3 compiled-object cps-compiled-field op)
   (let ((k (gen-kvar))
         (v-field (gen-var)))
      (lambda-cps k 
         (apply-cps cps-compiled-field 
            (lambda-cont v-field 
               (apply-cont k 
                  (make-Call FunUnaryOp 
                     (list (symbol-to-string-node op)
                        (make-Call FunAccess (list compiled-object v-field))))))))))

;; 4 - (not cps-ojbect) (not cps-field)
(define (compile-postfixed-op-access-aux4 compiled-object compiled-field op)
   (make-Call FunUnaryOp 
      (list (symbol-to-string-node op) 
         (make-Call FunAccess (list compiled-object compiled-field)))))

(define (compile-postfixed-op-ref env expr op)
   (let ((compiled-expr (cps-trans! expr env)))
      (if (is-a? compiled-expr Cps-Code)
         (let ((k (gen-kvar))
               (v (gen-var)))
            (lambda-cps k 
               (apply-cont compiled-expr 
                  (lambda-cont v
                     (apply-cont k 
                        (make-Call FunUnaryOp (list (symbol-to-string-node op) v)))))))  
         (make-Call FunUnaryOp (list (symbol-to-string-node op) compiled-expr))))) 

;;========== Pragma  =====================================================================================================================================
(define-nmethod (Pragma.cps-trans!) 
   (error "cps-trans!" "Finally I can see what is a Pragma node in Javascript"
      (class-name (object-class this))))

;;========== Literal  ====================================================================================================================================
(define-nmethod (Literal.cps-trans!) 
   this)

;;========== Ref =========================================================================================================================================
(define-nmethod (Ref.cps-trans!) 
   this)

;;========== This ========================================================================================================================================
(define-nmethod (This.cps-trans!)
   LocalThis)

;;========== NOP  ========================================================================================================================================
(define-nmethod (NOP.cps-trans!) 
   this)

;;========== Cond  =======================================================================================================================================
(define-nmethod (Cond.cps-trans!)
   (with-access::Cond this (test then else)
      (cond-cps-transform env test then else))) 

(define (cond-cps-transform env test then else) 
    (let ((k (gen-kvar))
          (v (gen-var))
          (compiled-test (cps-trans! test env))
          (compiled-then (cps-trans! then env))
          (compiled-else (cps-trans! else env))
          (cond-cont-var (make-Ref '_cond_body_cont)))
       (let ((cond-cont 
                (lambda-cont v 
                   (make-If v 
                      (if (is-a? compiled-then Cps-Code)
                         (apply-cps compiled-then k)
                         (make-Call k (list compiled-then))) 
                      (if (is-a? compiled-else Cps-Code)
                         (apply-cps compiled-else k)
                         (make-Call k (list compiled-else)))))))
          (lambda-cps k
             (list 
                (make-Var-Decl-List
                   (list (make-Init cond-cont-var cond-cont)))
                (if (is-a? compiled-test Cps-Code) 
                   (apply-cps compiled-test 
                      (lambda-cont v
                         (make-Call GetRealValue (list v cond-cont-var))))    
                (make-Call GetRealValue (list compiled-test cond-cont-var)))))))) 

;;========== New  ========================================================================================================================================
(define (my-acc fun lst acc-value)
   (if (not (pair? lst))
      acc-value
      (my-acc fun (cdr lst)  
         (apply fun (list (car lst) acc-value))))) 

(define-nmethod (New.cps-trans!) 
   (with-access::New this (class args)
      (let ((compiled-class (cps-trans! class env))
            (compiled-args 
               (map (lambda (arg) (cps-trans! arg env)) args)))
        (let ((are-normal-cps-args 
                 (my-acc (lambda (arg acc-value)
                            (or (is-a? arg Cps-Code) acc-value))
                         compiled-args
                         #f))
              (is-cps-class (is-a? compiled-class Cps-Code)))
          (cond 
             ((and are-normal-cps-args is-cps-class)
                 (let ((cps-compiled-args
                          (map (lambda (arg) (convert-simple-expr-to-cps arg)) compiled-args))) 
                   (compile-new-normal-cps-args-normal-cps-class compiled-class cps-compiled-args)))
             (are-normal-cps-args
                (let ((cps-compiled-args
                          (map (lambda (arg) (convert-simple-expr-to-cps arg)) compiled-args))) 
                   (compile-new-normal-cps-args-no-cps-class compiled-class cps-compiled-args)))
             (is-cps-class
                (compile-new-no-cps-args-normal-cps-class compiled-class compiled-args))
             (else (compile-new-no-cps-args-no-cps-class compiled-class compiled-args)))))))

(define (build-cps-args-call-chain compiled-args v index k)
   (if (pair? compiled-args)
      (lambda-cont (extend-var v index)
         (apply-cps (car compiled-args) 
            (build-cps-args-call-chain (cdr compiled-args) v (+ 1 index) k)))
      (lambda-cont (extend-var v index)
         (apply-cont k 
            (make-Call FunNew 
               (map (lambda (n) (extend-var v n)) (iota (+ index 1) 0 1)))))))

(define (compile-new-normal-cps-args-normal-cps-class compiled-class cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (lambda-cps k 
         (apply-cps compiled-class 
            (build-cps-args-call-chain cps-compiled-args v 0 k)))))

(define (compile-new-normal-cps-args-no-cps-class compiled-class cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (lambda-cps k 
         (apply-cps  
            (build-cps-args-call-chain cps-compiled-args v 0 k)
             compiled-class))))

(define (compile-new-no-cps-args-normal-cps-class compiled-class compiled-args)
   (let ((k (gen-kvar))
         (v-class (gen-var)))
      (lambda-cps k 
         (apply-cps compiled-class
             (lambda-cont v-class
                (apply-cont k 
                   (make-Call FunNew (cons v-class compiled-args))))))))

(define (compile-new-no-cps-args-no-cps-class compiled-class compiled-args)
   (make-Call FunNew (cons compiled-class compiled-args)))
     
;;========== Call  =======================================================================================================================================
(define-nmethod (Call.cps-trans!) 
   (with-access::Call this (fun args)
      (let ((compiled-args
               (map (lambda (arg) (cps-trans! arg env)) args)))
         (if (Access? fun)
            (with-access::Access fun (obj field)
               (let ((compiled-obj (cps-trans! obj env))
                     (compiled-field (cps-trans! field env)))
                  (cps-trans-method-call compiled-obj compiled-field compiled-args)))
            (cps-trans-function-call (cps-trans! fun env) compiled-args)))))

;;========== Function Call ===============================================================================================================================
(define (cps-trans-function-call compiled-fun compiled-args)
   (let ((are-normal-cps-args 
            (my-acc 
               (lambda (arg acc-value)
                  (or (is-a? arg Cps-Code) acc-value))
               compiled-args
               #f))
          (is-cps-fun (is-a? compiled-fun Cps-Code)))
      (cond  
         ;
         ((and are-normal-cps-args is-cps-fun)
             (let ((cps-compiled-args (map (lambda (arg) (convert-simple-expr-to-cps arg)) compiled-args))
                   (cps-compiled-fun (convert-simple-expr-to-cps compiled-fun)))
                (cps-trans-fun-call-aux1 cps-compiled-fun cps-compiled-args)))
         ;
         (are-normal-cps-args
             (let ((cps-compiled-args (map (lambda (arg) (convert-simple-expr-to-cps arg)) compiled-args)))
                (cps-trans-fun-call-aux2 compiled-fun cps-compiled-args)))
         ;
         (is-cps-fun (cps-trans-fun-call-aux3 compiled-fun compiled-args))
         ;
         (else (cps-trans-fun-call-aux4 compiled-fun compiled-args)))))

(define (cps-trans-fun-call-aux1 cps-compiled-fun cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (lambda-cps k 
         (apply-cps cps-compiled-fun 
             (function-call-build-cps-args-call-chain1 cps-compiled-args v 0 k)))))

(define (cps-trans-fun-call-aux2 compiled-function cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (lambda-cps k 
         (function-call-build-cps-args-call-chain2 cps-compiled-args v 1 k compiled-function))))

(define (cps-trans-fun-call-aux3 cps-compiled-fun compiled-args)
   (let ((k (gen-kvar))
         (v-f (gen-var)))
      (lambda-cps k 
         (apply-cps cps-compiled-fun 
            (lambda-cont v-f
               (make-Call FunFuncall (cons v-f (append compiled-args (list k)))))))))

(define (cps-trans-fun-call-aux4 compiled-fun compiled-args)
   (let ((k (gen-kvar)))
      (lambda-cps k 
         (make-Call FunFuncall 
            (cons compiled-fun (append compiled-args (list k)))))))

(define (function-call-build-cps-args-call-chain1 cps-compiled-args v index k)
   (if (pair? cps-compiled-args)
      (lambda-cont (extend-var v index)
         (apply-cps (car cps-compiled-args)
            (function-call-build-cps-args-call-chain1 (cdr cps-compiled-args) v (+ index 1) k)))
      (lambda-cont (extend-var v index) 
         (make-Call FunFuncall 
            (append 
               (map (lambda (n) (extend-var v n)) (iota (+ index 1) 0 1))
               (list k))))))

(define (function-call-build-cps-args-call-chain2 cps-compiled-args v index k compiled-function)
   (if (pair? cps-compiled-args)
      (apply-cps (car cps-compiled-args)
         (lambda-cont (extend-var v index)
            (function-call-build-cps-args-call-chain2 (cdr cps-compiled-args) v (+ index 1) k compiled-function)))
      (make-Call FunFuncall
         (cons compiled-function 
            (append 
               (map (lambda (n) (extend-var v n)) (iota (- index 1) 1 1))
               (list k))))))

;;========== Method Call ===============================================================================================================================
(define (cps-trans-method-call compiled-obj compiled-field compiled-args)
   (let ((are-cps-args 
            (my-acc 
               (lambda (arg acc-value)
                  (or (is-a? arg Cps-Code) acc-value))
               compiled-args
               #f))
         (is-cps-obj (is-a? compiled-obj Cps-Code))
         (is-cps-field (is-a? compiled-field Cps-Code)))
      (let ((cps-compiled-args (map (lambda (arg) (convert-simple-expr-to-cps arg)) compiled-args)))
         (cond  
            ;; cps-args cps-obj cps-field
            ((and are-cps-args is-cps-obj is-cps-field) 
               (cps-trans-method-call-aux1 compiled-obj compiled-field cps-compiled-args))
            ;; cps-args cps-obj (not cps-field)
            ((and are-cps-args is-cps-obj (not is-cps-field))
               (cps-trans-method-call-aux2 compiled-obj compiled-field cps-compiled-args))
            ;; cps-args (not cps-obj) cps-field
            ((and are-cps-args (not is-cps-obj) is-cps-field)
               (cps-trans-method-call-aux3 compiled-obj compiled-field cps-compiled-args))
            ;; cps-args (not cps-obj) (not cps-field)
            ((and are-cps-args (not is-cps-obj) (not is-cps-field))
               (cps-trans-method-call-aux4 compiled-obj compiled-field cps-compiled-args))
            ;; (not cps-args) cps-obj cps-field
            ((and (not are-cps-args) is-cps-obj is-cps-field)
               (cps-trans-method-call-aux5 compiled-obj compiled-field compiled-args))
            ;; (not cps-args) cps-obj (not cps-field)
            ((and (not are-cps-args) is-cps-obj (not is-cps-field))
               (cps-trans-method-call-aux6 compiled-obj compiled-field compiled-args))
            ;; (not cps-args) (not cps-obj) cps-field
            ((and (not are-cps-args) (not is-cps-obj) is-cps-field)
               (cps-trans-method-call-aux7 compiled-obj compiled-field compiled-args))
            ;; (not cps-args) (not cps-obj) (not cps-field)
            ((and (not are-cps-args) (not is-cps-obj) (not is-cps-field))
               (cps-trans-method-call-aux8 compiled-obj compiled-field compiled-args))))))

;; 1 - cps-args cps-obj cps-field
(define (cps-trans-method-call-aux1 cps-compiled-object cps-compiled-field cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-o (extend-var v "o"))
            (v-f (extend-var v "f"))
            (v-arg (extend-var v "arg")))
         (lambda-cps k 
            (apply-cps cps-compiled-object 
               (lambda-cont v-o
                  (apply-cps cps-compiled-field
                     (lambda-cont v-f
                        (method-call-build-cps-args-call-chain cps-compiled-args v-arg 1 v-o v-f k)))))))))

;; 2 - cps-args cps-obj (not cps-field)
(define (cps-trans-method-call-aux2 cps-compiled-object compiled-field cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-o (extend-var v "o"))
            (v-arg (extend-var v "arg")))
         (lambda-cps k 
            (apply-cps cps-compiled-object
               (lambda-cont v-o 
                  (method-call-build-cps-args-call-chain cps-compiled-args v-arg 1 v-o compiled-field k)))))))

;; 3 - cps-args (not cps-obj) cps-field
(define (cps-trans-method-call-aux3 compiled-object cps-compiled-field cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-f (extend-var v "f"))
            (v-arg (extend-var v "arg")))
         (lambda-cps k 
            (apply-cps cps-compiled-field
               (lambda-cont v-f 
                  (method-call-build-cps-args-call-chain cps-compiled-args v-arg 1 compiled-object v-f k)))))))

;; 4 - cps-args (not cps-obj) (not cps-field)
(define (cps-trans-method-call-aux4 compiled-object compiled-field cps-compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-arg (extend-var v "arg")))
         (lambda-cps k 
            (method-call-build-cps-args-call-chain cps-compiled-args v-arg 1 compiled-object compiled-field k)))))

;; 5 - (not cps-args) cps-obj cps-field
(define (cps-trans-method-call-aux5 cps-compiled-object cps-compiled-field compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-o (extend-var v "o"))
            (v-f (extend-var v "f")))
         (lambda-cps k
            (apply-cps cps-compiled-object 
               (lambda-cont v-o 
                  (apply-cps cps-compiled-field
                     (lambda-cont v-f 
                        (make-Call FunMethodcall
                           (cons v-o (cons v-f (append compiled-args (list k)))))))))))))

;; 6 - (not cps-args) cps-obj (not cps-field)
(define (cps-trans-method-call-aux6 cps-compiled-object compiled-field compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-o (extend-var v "o")))
         (lambda-cps k
            (apply-cps cps-compiled-object 
               (lambda-cont v-o
                  (make-Call FunMethodcall
                     (cons v-o (cons compiled-field (append compiled-args (list k)))))))))))

;; 7 - (not cps-args) (not cps-obj) cps-field
(define (cps-trans-method-call-aux7 compiled-object cps-compiled-field compiled-args)
   (let ((k (gen-kvar))
         (v (gen-var)))
      (let ((v-f (extend-var v "f")))
         (lambda-cps k
            (apply-cps cps-compiled-field 
               (lambda-cont v-f
                  (make-Call FunMethodcall
                     (cons compiled-object (cons v-f (append compiled-args (list k)))))))))))

;; 8 - (not cps-args) (not cps-obj) (not cps-field)
(define (cps-trans-method-call-aux8 compiled-object compiled-field compiled-args)
   (let ((k (gen-kvar)))
      (lambda-cps k
         (make-Call FunMethodcall
            (cons compiled-object (cons compiled-field (append compiled-args (list k))))))))

(define (method-call-build-cps-args-call-chain cps-compiled-args v index object field k)
   (if (pair? cps-compiled-args)
      (apply-cps (car cps-compiled-args)
         (lambda-cont (extend-var v index)
            (method-call-build-cps-args-call-chain (cdr cps-compiled-args) v (+ index 1) object field k)))
      (make-Call FunMethodcall    
         (cons object (cons field 
            (append 
               (map (lambda (n) (extend-var v n)) (iota (- index 1) 1 1))
               (list k))))))) 

;;========== Array =======================================================================================================================================
(define (my-map fun lst)
   (define (my-map-aux fun lst index lst-ret)
      (if (pair? lst) 
         (my-map-aux fun (cdr lst) (+ index 1)
            (cons (apply fun (list (car lst) index)) lst-ret))
         (reverse lst-ret)))
   (my-map-aux fun lst 0 (list)))

(define-nmethod (Array.cps-trans!)
   (with-access::Array this (els len)
      (let ((compiled-array-els
               (map 
                  (lambda (el) (cps-trans! (Array-Element-expr el) env))
                  els)))
         (let ((is-any-array-el-cps 
                  (my-acc 
                     (lambda (arg acc-value)
                        (or (is-a? arg Cps-Code) acc-value))
                     compiled-array-els
                     #f)))
             (if is-any-array-el-cps
                (let ((cps-compiled-array-els 
                         (map (lambda (el) 
                                 (convert-simple-expr-to-cps el))
                            compiled-array-els))
                      (k (gen-kvar))
                      (v (gen-var)))
                   (lambda-cps k 
                      (cps-trans-array-els cps-compiled-array-els v 0 k)))
                (make-Array 
                   (my-map 
                      (lambda (el index) 
                         (make-Array-Element index el))
                      compiled-array-els)
                   len))))))

(define (cps-trans-array-els cps-compiled-array-els v index k)
   (if (pair? cps-compiled-array-els)
      (apply-cps (car cps-compiled-array-els)
         (lambda-cont (extend-var v index)
            (cps-trans-array-els (cdr cps-compiled-array-els) v (+ index 1) k)))
      (apply-cont k 
         (make-Array (map (lambda (n) (make-Array-Element n (extend-var v n)))
                        (iota index 0 1))
            index))))
                       
;;========== Obj-Init  ===================================================================================================================================
(define (my-map-double fun lst1 lst2)
   (cond ((and (pair? lst1) (pair? lst2)) 
             (cons (apply fun (list (car lst1) (car lst2)))
                   (my-map-double fun (cdr lst1) (cdr lst2))))
         ((pair? lst1) #f)
         ((pair? lst2) #f)
         (else (list))))

(define-nmethod (Obj-Init.cps-trans!) 
   (with-access::Obj-Init this (inits)
      (let ((props 
               (map (lambda (el) 
                       (with-access::Property-Init el (name val)
                          name))
                  inits))
            (compiled-prop-values 
               (map (lambda (el)
                       (with-access::Property-Init el (name val)
                          (cps-trans! val env)))
                  inits)))
         (let ((is-any-prop-value-cps
                  (my-acc 
                     (lambda (arg acc-value)
                        (or (is-a? arg Cps-Code) acc-value))
                     compiled-prop-values
                     #f)))
            (if is-any-prop-value-cps 
               (let ((cps-compiled-prop-values
                        (map (lambda (el) 
                                (convert-simple-expr-to-cps el))
                           compiled-prop-values))
                     (k (gen-kvar))
                     (v (gen-var)))
                  (lambda-cps k 
                     (cps-trans-inits cps-compiled-prop-values v 0 k props)))
               (make-Obj-Init 
                  (my-map-double 
                     (lambda (prop prop-value)
                        (make-Property-Init prop prop-value))
                     props
                     compiled-prop-values)))))))

(define (cps-trans-inits cps-compiled-prop-values v index k props)
   (if (pair? cps-compiled-prop-values) 
       (apply-cps (car cps-compiled-prop-values) 
          (lambda-cont (extend-var v index)
             (cps-trans-inits (cdr cps-compiled-prop-values) v (+ index 1) k props)))
       (apply-cont k 
          (make-Obj-Init 
             (my-map 
                (lambda (prop index)
                   (make-Property-Init prop (extend-var v index)))
                props))))) 

;;========== Named-Fun ===================================================================================================================================
(define-nmethod (Named-Fun.cps-trans!)
   (with-access::Named-Fun this (name fun)
      (make-Named-Fun name 
         (cps-trans! fun env))))

;;========== Fun  ========================================================================================================================================
(define-nmethod (Var-Fun.cps-trans!) 
   (with-access::Var-Fun this (params body var-decl)
      (let ((k (gen-kvar))
            (v (gen-var)))
         (let ((new-env (return-env k)))
            (let ((compiled-body (cps-trans! body new-env)))
               (make-Var-Fun 
                  (cons k params)
                  (if (is-a? compiled-body Cps-Code)
                     (make-Block (list (apply-cps compiled-body k))) 
                     (my-make-block
                        (list 
                           compiled-body
                           (apply-cps k (make-Undefined 0)))))
                  var-decl))))))

;;========== Sequence  ===================================================================================================================================
(define-nmethod (Sequence.cps-trans!)
   (with-access::Sequence this (exprs)
      (let ((k (gen-kvar))
            (compiled-exprs 
               (map 
                  (lambda (expr) (cps-trans! expr env))
                  exprs)))
         (let ((cps-compiled-exprs 
                  (process-sequenced-compiled-exprs compiled-exprs)))
            (let ((compiled-sequence 
                     (cond 
                        ((pair? cps-compiled-exprs)
                            (lambda-cps k (cps-trans-exprs cps-compiled-exprs k)))
                        ((not (null? cps-compiled-exprs)) cps-compiled-exprs)
                        (else (make-Undefined 0)))))
               compiled-sequence)))))


(define (cps-trans-exprs exprs k) 
   (let ((v (gen-var)))
      (cond ((pair? (cdr exprs)) 
               (apply-cps (car exprs)
                  (lambda-cont v (cps-trans-exprs (cdr exprs) k))))
            ((pair? exprs)
               (apply-cps (car exprs) k))
            (else (error "cps-trans!" "Missing case in cps-trans-exprs." #f)))))

;; how to deal with blocks/ sequences of expressions...
(define (process-sequenced-compiled-exprs compiled-exprs)
   (define (process-sequenced-compiled-exprs compiled-exprs cps-compiled-exprs acc-non-cps-compiled-exprs)
      (cond ((and (pair? compiled-exprs) (pair? acc-non-cps-compiled-exprs) (is-a? (car compiled-exprs) Cps-Code))
               (let ((updated-current-cps-term 
                        (update-cps-term-with-non-cps-code (car compiled-exprs) (process-sequence-of-non-cps-exprs acc-non-cps-compiled-exprs))))
                  (process-sequenced-compiled-exprs 
                     (cdr compiled-exprs)
                     (append cps-compiled-exprs (list updated-current-cps-term))
                     (list))))
            ((and (pair? compiled-exprs) (not (pair? acc-non-cps-compiled-exprs)) (is-a? (car compiled-exprs) Cps-Code))
               (process-sequenced-compiled-exprs 
                      (cdr compiled-exprs)
                      (append cps-compiled-exprs (list (car compiled-exprs)))
                      (list)))
            ((and (pair? compiled-exprs) (not (is-a? (car compiled-exprs) Cps-Code)))
               (process-sequenced-compiled-exprs 
                  (cdr compiled-exprs) 
                  cps-compiled-exprs
                  (append acc-non-cps-compiled-exprs (list (car compiled-exprs)))))
            ((and (not (pair? compiled-exprs)) (pair? acc-non-cps-compiled-exprs) (not (pair? cps-compiled-exprs)))
               (if (> (length acc-non-cps-compiled-exprs) 1) 
                  (let ((grouped-non-cps-exprs (group-non-cps-exprs acc-non-cps-compiled-exprs)))
                     grouped-non-cps-exprs)
                  (car acc-non-cps-compiled-exprs)))
            ((and (not (pair? compiled-exprs)) (pair? acc-non-cps-compiled-exprs))
               (let ((cps-grouped-non-cps-exprs (cps-group-non-cps-exprs acc-non-cps-compiled-exprs)))
                  (append cps-compiled-exprs (list cps-grouped-non-cps-exprs))))
            ((and (not (pair? compiled-exprs)) (not (pair? acc-non-cps-compiled-exprs)))
               (if (eq? (length cps-compiled-exprs) 1)
                  (car cps-compiled-exprs)
                  cps-compiled-exprs))))
   (process-sequenced-compiled-exprs 
      (merge-consecutive-blocks (remove-undefineds-from-sequence-of-exprs compiled-exprs))
      (list) 
      (list)))

(define (does-list-contain-blocks lst-exprs)
   (cond ((null? lst-exprs) #f)
         ((is-a? (car lst-exprs) Block) #t)
         (else (does-list-contain-blocks (cdr lst-exprs)))))

(define (group-non-cps-exprs lst-exprs)
   (let ((number-of-exprs (length lst-exprs)))
      (cond 
         ((eq? number-of-exprs 0) (make-Undefined 0))
         ((eq? number-of-exprs 1) (car lst-exprs))
         (else 
            (make-Block (process-sequence-of-non-cps-exprs lst-exprs))))))

(define (process-sequence-of-non-cps-exprs lst-exprs)
  (if (does-list-contain-blocks lst-exprs)
     (let loop ((lst-exprs-aux lst-exprs)
                (simple-exprs (list)))
        (if (null? lst-exprs-aux) 
           simple-exprs
           (loop 
              (cdr lst-exprs-aux)
              (if (is-a? (car lst-exprs-aux) Block)
                 (with-access::Block (car lst-exprs-aux) (stmts)
                    (append simple-exprs stmts))
                 (append simple-exprs (list (car lst-exprs-aux)))))))
     lst-exprs))     

(define (cps-group-non-cps-exprs non-cps-exprs)
   (let* ((k (gen-kvar))
          (exprs (append 
                    (process-sequence-of-non-cps-exprs non-cps-exprs) 
                    (list (apply-cont k (make-Undefined 0))))))
      (lambda-cps k 
         (make-Block exprs))))

(define (update-cps-term-with-non-cps-code cps-term list-of-non-cps-exprs)
   (with-access::Cps-Code cps-term (params body)
      (lambda-cps params
         (if (is-a? body Block)
            (with-access::Block body (stmts)
               (append list-of-non-cps-exprs stmts))
            (append list-of-non-cps-exprs (list body))))))

;; printing code
(define (print-list-of-expressions lst)
   (print "Printing list of expressions")
   (let loop ((lst-aux lst)
              (i 1)) 
      (if (pair? lst-aux)
         (begin 
            (print (format "expression:~a" i))
            (print (class-name (object-class (car lst-aux))))
            (js-out (car lst-aux) (current-error-port) :compress? #f)
            (loop (cdr lst-aux) (+ i 1))))))

;; merge consecutive blocks code
(define (merge-consecutive-blocks lst)
   (let loop ((exprs-yet-to-analyse lst)
              (analysed-exprs (list)))
      (if (null? exprs-yet-to-analyse)
         analysed-exprs
         (if (is-a? (car exprs-yet-to-analyse) Block)
            (with-access::Block (car exprs-yet-to-analyse) (stmts) 
               (loop 
                  (cdr exprs-yet-to-analyse) 
                  (append analysed-exprs stmts)))
            (loop 
               (cdr exprs-yet-to-analyse)
               (append analysed-exprs (list (car exprs-yet-to-analyse))))))))

;; remove undefineds      
;; florian's parser error - does not parse undefined - parse undefined as ref!!!!!!!!!
(define (remove-undefineds-from-sequence-of-exprs compiled-exprs)
   (let loop ((lst-exprs compiled-exprs)
              (lst-exprs-output (list)))
      (if (null? lst-exprs) 
         (reverse lst-exprs-output)
         (if (is-a? (car lst-exprs) Undefined)
            (begin 
               (loop (cdr lst-exprs) lst-exprs-output))
            (begin
               ;;(print (class-name (object-class (car lst-exprs))))
               (loop (cdr lst-exprs) (cons (car lst-exprs) lst-exprs-output)))))))

;;========== Block  ======================================================================================================================================
(define-nmethod (Block.cps-trans!) 
   (let ((out-p (current-error-port)))
      (with-access::Block this (stmts)
         (let ((compiled-stmts 
                  (map 
                     (lambda (stmt) (cps-trans! stmt env))
                     stmts)))
            (let ((cps-compiled-stmts 
                    (process-sequenced-compiled-exprs compiled-stmts)))
               (if (pair? cps-compiled-stmts)
                  (let* ((k (gen-kvar))
                         (compiled-block (cps-trans-stmts cps-compiled-stmts k)))
                     (lambda-cps k
                        compiled-block))
                  cps-compiled-stmts))))))

(define (cps-trans-stmts stmts k)
   (cond ((pair? (cdr stmts))  
            (apply-cps (car stmts)
               (lambda-cont dummy-var (cps-trans-stmts (cdr stmts) k))))
         ((pair? stmts)
            (apply-cps (car stmts) k))
         (else (error "cpstrans!" "cannot call cps-trans-stmts with a single stmt" #f))))