(module js2post-main
   (import config
           js2post)
   (main my-main))

(define *in-file* #f)
(define *out-file* #f)
(define *trampolines* #f)
(define *var-list* '()); a list of foreign names imported by script (e.g. the google object)

(define *version* "20120725")

(define (handle-args args)
   (args-parse (cdr args)
      (section "Help")
      (("?")
       (args-parse-usage #f)
       (exit))
      ((("-h" "--help") (help "?,-h,--help" "This help message"))
       (args-parse-usage #f)
       (exit))
      (section "Misc")
      ((("--version") (help "Version number"))
       (print *version*)
       (exit))
      ((("-v" "--verbose") (help "Verbose output"))
       (config-set! 'verbose #t))
      (("-o" ?file (help "The output file. '-' prints to stdout."))
       (set! *out-file* file))
      (("-i" ?file (help "The input file. '-' reads from stdin."))
       (set! *in-file* file))
      (("-t" (help "Trampolined transformation"))
       (set! *trampolines* #t)) 
      (("-a" ?varl ?vart ?fid (help "'VARL' should point to 'VART' imported from inner iframe and FID should be the frame where this gadget object is!!!"))
       (set! *var-list* (cons (list (string->symbol varl) (string->symbol vart) (string->symbol fid)) *var-list*)))
      (else
       (begin
          (display "Unknow argument: " (current-error-port))
          (display else (current-error-port))
          (args-parse-usage #f)
          (exit)))))

(define (my-main args)
   (config-init!)
   (handle-args args)
   (let ((in-p (if (not *in-file*) (current-input-port) (open-input-file *in-file*)))
         (out-p (if (not *out-file*) (current-output-port) (open-output-file *out-file*))))
      (with-handler
         (lambda (e)
            (unless (not *out-file*) (delete-file *out-file*))
            (raise e))
         (unwind-protect
            (js2post-compile-file in-p out-p *var-list* *trampolines*)
            (begin (close-input-port in-p) (close-output-port out-p))))
      0))

(define (print-tuples-of-three lst) 
   (unless (null? lst)
      (let ((first (caar lst))
            (second (cadar lst))
            (third (caddar lst)))
         (print (format "first: ~a, second: ~a, third: ~a" first second third)))
      (print-tuples-of-three (cdr lst))))

