/// Useful auxiliary functions (used to be common.js)
// verbose(...) uses console to log messages.
var turn_on_verbose;
var turn_off_verbose;
var verbose;

var tramp_on = true;

(function(){
    var verbose_flag = false;
    turn_on_verbose = function(){ verbose_flag = true; };
    turn_off_verbose = function(){ verbose_flag = false; };
    verbose = function(){
	if (!verbose_flag)
	    return;
	var log = new String();
	for (var i = 0; i < arguments.length; i++){
	    log += arguments[i];
	}
	if (arguments.length > 0)
	    console.log(log);
	return;
    };
})();

//Turn on verbose
turn_on_verbose();

// handle_id_gen() generate a fresh interger id for various usage.
var handle_id_gen = (function(){
	     var id_counter = 0;
		 return (function(){return id_counter++;});})();

// msg_id_gen() generate a fresh interger id for various usage.
var msg_id_gen = (function(){
	     var id_counter = 0;
		 return (function(){return id_counter++;});})();



//Data types of object handler of iframe pages
function OHandle(){
    var id = handle_id_gen();
    this.__id = id;
    this.__is_ohandle = true;
    return;
}

//Data types of funcation call back handlers.
function FHandle(){
    var id = handle_id_gen();    
    this.__id = id;
    this.__is_fhandle = true;
    return;
}

var startTime;

function timerStart(){
    startTime = (new Date()).getTime();
}

function timerEndInternal(){
    var endTime = (new Date()).getTime();
    var time = endTime - startTime;
    var logDiv = document.getElementById("log");
    logDiv.innerHTML="Consumed time: " + time + "ms.";
}

function timerEnd(){
    var t = setTimeout("timerEndInternal()",00);
}

/// Trampoline
function __Tramp(fun,arg){
    this.fun=fun;
    this.arg=arg;
    this.restart=function(){
	return this.fun(this.arg);
    }
}

var tramp_in_flag=false;

function execute_as_tramp(fun,arg){
    var _res_or_tramp;
    _c_g = 0;
    tramp_in_flag=true;
    _res_or_tramp=fun(arg);
    while (_res_or_tramp instanceof __Tramp)
	_res_or_tramp = _res_or_tramp.restart();
    tramp_in_flag=false;
}

function callback_as_tramp(fun,args){
    var _res_or_tramp;
    _c_g = 0;
    tramp_in_flag=true;
    _res_or_tramp=fun.apply(null,args);
    while (_res_or_tramp instanceof __Tramp)
	_res_or_tramp = _res_or_tramp.restart();
    tramp_in_flag=false;
}

/// Proxy Library

//parameters
var untrusted_origin="*";
var untrusted_iframe=false;
var untrusted_iframe_id='iframe';

//Initailize the list for storing continuations and callback functions
var set_cont;
var get_cont;
var get_src;
var set_callback;
var get_callback;

(function(){
    var cont_list = new Object();
    var src_list = new Object();
    var callback_list = new Object();

    //Set continuations
    set_cont = function(frame_id, msg_id,cont){
	cont_list[msg_id.toString()] = cont;
	src_list[msg_id.toString()] = frame_id; 
    }
    
    //Get continuations
    get_cont = function(msg_id){
	return cont_list[msg_id.toString()];
    }

    get_src = function(msg_id){
	return document.getElementById(src_list[msg_id.toString()]).contentWindow;
    }
    
    //Set callbacks
    set_callback = function(fhandle,callback){
	cont_list[fhandle.__id.toString()] = callback;
    }
    
    //Get continuations
    get_callback = function(fhandle){
	return cont_list[fhandle.__id.toString()];
    }
})();

function message_handler(e){
  var recv = JSON.parse(e.data);
  var val;
  switch(recv.msg_type){
      //execute continuation	
  case 'EXE_CONT':
      var cont = get_cont(recv.msg_id);
      var src_window = get_src(recv.msg_id);
      if (cont!=null && (e.source == src_window)){
	  set_cont("", recv.msg_id,null);
	  if (cont.val!=undefined){
	      val = cont.val;
	  }else{
	      val = recv.return_val;
	  }
	  if (tramp_on)
	      execute_as_tramp(cont,val);
	  else
	      cont(recv.val);
      }
      else{
	  throw "Null continuation";
      }
      break;
      //execute callback function
  case 'CALLBACK':
      var callback = get_callback(recv._function);
      if (callback!=null){
	  if (tramp_on)
	      callback_as_tramp(callback,recv.arguments);
	  else
	      callback.apply(null,recv.arguments);
      }
      else{
	  throw "Null callback";
      }
      break;
  }
}

//init the message handler
(function(){
    window.addEventListener("message", message_handler, false);
})();

// function val_recon(val){
//     return val;
// }

// function args_recon(args){
//     return args;
// }

function postMsg(message, frame_id){
    if (frame_id == undefined){
	if (!untrusted_iframe)
	    untrusted_iframe = document.getElementById(untrusted_iframe_id).contentWindow; 
    	json_message =  JSON.stringify(message);
	verbose("O=>I Request"/*,untrusted_origin */, " : ", json_message);
	untrusted_iframe.postMessage(json_message,untrusted_origin);
    } else {
	var target_iframe = document.getElementById(frame_id).contentWindow;
	json_message =  JSON.stringify(message);
	verbose("O=>I Request"/*,untrusted_origin */, " : ", json_message);
	target_iframe.postMessage(json_message,untrusted_origin);
    }
}

function GET_GLOBAL_REF(frame_id,global_name,cont){
    
    //send message
    var m_id = msg_id_gen();
    var msg =  {msg_id : m_id, msg_type : 'GET_GLOBAL_REF', global_name : global_name};
    postMsg(msg, frame_id);
    
    //register continuation
    set_cont(frame_id, m_id,cont);
    return;
}

function GET_PROPS_FOR_IN(frame_id, obj, cont){
    //send message
    var m_id = msg_id_gen();
    var msg =  {msg_id : m_id, msg_type: 'GET_PROPS_FOR_IN', object : obj};
    postMsg(msg, frame_id);
    
    //register continuation
    set_cont(frame_id, m_id,cont);
    return;
}

// CHANGE TO GET_PROPERTY_REAL; ADDING FLUSHING
function GET_PROPERTY(frame_id,obj,prop,cont){
    //send message
    var m_id = msg_id_gen();
    var msg =  {msg_id : m_id, msg_type: 'GET_PROPERTY', object : obj, property_name : prop};
    postMsg(msg, frame_id);
    
    //register continuation
    set_cont(frame_id, m_id,cont);
    return;
}

function OBJ_PROP_ASSIGN(frame_id,obj,prop,val,cont){
    //send message
    var m_id = msg_id_gen();
    var send_val;
    if (typeof val == 'function'){
	send_val = new FHandle();
	set_callback(send_val,val);
    }else{
	send_val=val;
    }
    var msg =  {msg_id : m_id, msg_type: 'OBJ_PROP_ASSIGN', object : obj, property_name : prop, value: send_val};
    postMsg(msg, frame_id);
    
    //NEED TO REFINE THIS
    cont.val=val;
    
    //register continuationn
    set_cont(frame_id, m_id,cont);
    return;
}

function isSimpleObject(v){
	if (typeof v == 'object')
	   for (prop in v)
	      if (typeof v[prop] == 'object' || typeof v[prop] == 'function')
		return false;		
        return true;
}

function construct_args(offset, __arguments){
    var args = new Array();
    for (var i= 0 + offset; i<__arguments.length; i++){
	if (typeof __arguments[i] == 'function'){
	    args[i - offset] = new FHandle();
	    set_callback(args[i - offset],__arguments[i]);
	}
	else{
	    args[i - offset] = __arguments[i];
	}
    }
    return args;
}

function construct_args_untouch(offset, __arguments){
    var args = new Array();
    for (var i= 0 + offset; i<__arguments.length; i++){
	args[i - offset] = __arguments[i];
    }
    return args;
}

function CALL_FUNCTION(frame_id,fun,cont,args){
   //send message
    var m_id = msg_id_gen();

    // var offset =2;
    // var args = construct_args(offset,arguments);
    
    var msg =  {msg_id : mid, msg_type : 'CALL_FUNCTION', _function : fun, arguments : args};
    postMsg(msg, frame_id);
    
    //register continuation
    set_cont(frame_id, mid,cont);
    return;
}

function CALL_METHOD(frame_id,obj,method,cont,args){
   //send message
    var m_id = msg_id_gen();

    //var offset = 3;
    //var args = construct_args(offset,arguments);
    
    var msg =  {msg_id : m_id, msg_type: 'CALL_METHOD', object : obj, method_name: method, arguments : args};
    postMsg(msg, frame_id);

    //register continuation
    set_cont(frame_id, m_id,cont);
    return;
}

function NEW_OBJECT(frame_id,fun,cont,args){
   //send message
    var m_id = msg_id_gen();
   
    // var offset =2;
    // var args = construct_args(offset,arguments);
    
    var msg =  {msg_id : m_id, msg_type : 'NEW_OBJECT', _function : fun, arguments : args};
    postMsg(msg, frame_id);
    
    //register continuation
    set_cont(frame_id, m_id,cont);
    return;
}


function __ACCESS(frame_id,kont,obj,prop){
    if (obj.__is_ohandle)
	return GET_PROPERTY(frame_id,obj,prop,kont);
    else
	return kont(obj[prop]);
}

function __GETPROPS(frame_id, kont, obj){
    if (obj.__is_ohandle)
	return GET_PROPS_FOR_IN(frame_id,obj,kont);
    else{
	var props = new Array();
	for (prop in obj)
	    props.push(prop);
	return kont(props);
    }
       

}

function __METHODCALL(frame_id, kont,object,method){
    var fun,args;
    var offset = 4;

    if (object.__is_ohandle){
	args = construct_args(offset,arguments);
	return CALL_METHOD(frame_id,object,method,kont,args);
    }else{//NEED TO TEST IF THE FUNCTION IS CPS STYLE OR NOT
	fun=object[method];
	args=construct_args_untouch(offset,arguments);
	if(fun.__is_cps){	    
	    args.unshift(kont);
	    return fun.apply(object,args);
	}else
	    return kont(fun.apply(object,args));					    
    }
}

function __FUNCALL(frame_id, kont,object){
    var fun;
    
    var offset =3;
    var args;
    
    if (object.__is_ohandle){
	args = construct_args(offset,arguments);
	return CALL_FUNCTION(frame_id,object,kont,args);
    }else{//NEED TO TEST IF THE FUNCTION IS CPS STYLE OR NOT
	fun=object;
	args = construct_args_untouch(offset,arguments);
	if(fun.__is_cps){	    
	    args.unshift(kont);
	    return fun.apply(null,args);
	} else
	    return kont(fun.apply(null,args));					    
    }    
}

function __NEW(frame_id,kont,fun){
    var offset =3;
    var args = construct_args(offset,arguments);

    var val;

    var new_kont;

    if(fun.__is_ohandle){
	return NEW_OBJECT(frame_id,fun,kont,args);
    }else{
	if(!fun.__is_cps){
	    if (args.length == 0)
		val = new fun();
	    else{
		var expr = new String;
		expr = "new fun(";
		for (var i = 0; i<args.length - 1; i++)
		    expr = expr + "args[" + i + "],";
		expr  = expr + "args[" + i + "])";
		val = eval(expr);
	    }	
	}else{
	    new_kont=(
		      function(v){
			  return kont(val);
		      });
	    if (args.length == 0)
		return (val = new fun(new_kont));
	    else{
		var expr = new String;
		expr = "new fun(new_kont,";
		for (var i = 0; i<args.length - 1; i++)
		    expr = expr + "args[" + i + "],";
		expr  = expr + "args[" + i + "])";
		return (val = eval(expr));
	    }
	}
	return kont(val);
    }
}

function __GETREF(frame_id,global_name,cont){
    GET_GLOBAL_REF(frame_id,global_name,cont);
}

function __ACCSIG(frame_id,kont,obj,prop,val){
    if (obj.__is_ohandle && obj.hasOwnProperty('__is_ohandle')){
	return OBJ_PROP_ASSIGN(frame_id,obj,prop,val,kont);
    }else{
	obj[prop]=val;
	return kont(val);
    }
}

function __MAKEFUN(fun){    //CAUTION: need a better way to test if it is supplied with continuations!
    var returnfun = function(){
	//var offset = 0;
	//var args = construct_args(offset,arguments);
	var args = new Array();
	//verbose("Type of first argument is: ", typeof(arguments[0]));
	for (var i=0; i<arguments.length; i++)
	    args[i]=arguments[i];
	if (typeof(args[0]) != 'function'){
	    args.unshift(function(){});
	}
	
	if(tramp_in_flag)
	    return fun.apply(this,args);
	else{
	    var _res_or_tramp;
	    _c_g = 0;
	    tramp_in_flag=true;
	    _res_or_tramp=fun.apply(this,args);
	    while (_res_or_tramp instanceof __Tramp)
		_res_or_tramp = _res_or_tramp.restart();
	    tramp_in_flag=false;
	    return _res_or_tramp;
	}

    };
    returnfun.__is_cps=true;
    return returnfun;
}